/***
 * GASP: a gaming services platform for mobile multiplayer games.
 * MooDS: a communication protocol used in GASP
 *
 * Copyright (C) 2004,2005 CNAM/INT - 2006 CNAM/INT/Filao
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Romain Pellerin, Chanty Lim
 */
package org.mega.moods.generator;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;


/**
 * MooDS Generator v2.0
 * @author Romain PELLERIN
 */
public class CustomTypesGenerator {

    private static Vector classLanguages = new Vector();
    private static Vector classFiles = new Vector();
    private static Vector classNames = new Vector();
    private static Hashtable stringsToWrite = new Hashtable();
    public static boolean isMulticastCommunicationModelMethodsToAdd = false;
    public static boolean isUnicastCommunicationModelMethodsToAdd = false;
    
    // languages:
    // 0: Java
    // 1: Flash
    public static final int JAVA = 0;
    public static final int FLASH = 1;
    
    // t: strings array to write
    // t[0]: packages
    // t[1]: imports
    // t[2]: class declaration
	// t[3]: public attributes
    // t[4]: constructor
    // t[5]: encodeData
    // t[6]: decodeData
    // t[7]: encodingMethods
    // t[8]: decodingMethods
    // t[9]: string encoding utils methods
    // t[10]: communication methods (multicast)
    // t[11]: communication methods (unicast)
    // t[12]: end of class
  
    public static void init(){
    	classFiles = new Vector();
    	classLanguages = new Vector();
    	classNames = new Vector();
    	stringsToWrite = new Hashtable();
    	
    }
    
    
    /* TODO: reste  gnrer le serveur grant des hashtables et non des objets */
    public static void startGenerateCustomTypes(int language, int encoding, String _package) throws IOException{

    	String[] strings = new String[]{"","","","","","","","","","","","",""};
    	String path = Generator.getPath(language,false);
    	String className = (Generator.isClient? "CustomTypes":"CustomTypes"+((encoding == Generator.STRING)? "4StringEncoding":""));
    	String linkClass = path+className;
    	ClassMaker.createClassEntry(language,linkClass,className,strings,false);

	    classNames.add(linkClass);
    	classLanguages.add(""+language);
	    
    	// strings to write table for that class

    	// init strings table
    	strings[0] = "/** \r"
			+" * MooDS Encoder/Decoder for Flash.\r"
			+" * Encoding version: "+Generator.ENCODING_VERSION+"\r"
	 		+" * @author Generated by MooDS Generator v"+Generator.MOODS_VERSION+" - "+Generator.MOODS_YEAR+" "+Generator.MOODS_TEAM+" \r"
	 		+" */\r\r";
    	strings[0] += LanguageTranslator.writePackage(language,_package+"."+(Generator.isClient? "client"+((language == FLASH)? ".flash":""):"server"))+"\r";
        
    	// required imports
        if (language == FLASH)
        	strings[1] += LanguageTranslator.writeImport(language, "Hashtable.as")+"\r\r";
	    
        if (language == JAVA){
	    	strings[1] += LanguageTranslator.writeImport(language,  _package+".*")+"\r";
	    	strings[1] += LanguageTranslator.writeImport(language,"java.util.Hashtable")+"\r"; 
	    	if (encoding == Generator.BINARY){
	    		strings[1] += LanguageTranslator.writeImport(language, "java.io.DataOutputStream")+"\r";
	    		strings[1] += LanguageTranslator.writeImport(language, "java.io.DataInputStream")+"\r";	
        	}
        }

    	// class declaration
        strings[2] += LanguageTranslator.writeComments(language, "This class has been generated by MooDS Generator to handle "+Generator.ioApproach+"\r// for binarisation input/output.")+"\r";
    	if (Generator.isClient) strings[2] += LanguageTranslator.writeClassDeclaration(language,"CustomTypes",false,false,"")+"\r"; 
		else{ //server
		    	if (encoding == Generator.BINARY) strings[2] += LanguageTranslator.writeClassDeclaration(language,"CustomTypes",true,false,"org.mega.gasp.moods.CustomTypes")+"\r";
		    	if (encoding == Generator.STRING) strings[2] += LanguageTranslator.writeClassDeclaration(language,"CustomTypes4StringEncoding",true,false,"org.mega.gasp.moods.string.CustomTypes")+"\r";  	
		}
     	 	
    	
    	// start to generate encodeData root method
     	strings[5] += startToWriteEncodeDataMethod(language, encoding);
    	
     	// start to generate decodeData root method
     	strings[6] += startToWriteDecodeDataMethod(language, encoding);
    	
    	
    	if (encoding == Generator.STRING) strings[9] += addStringEncodingSpecificMethods(language);
    	strings[12] += LanguageTranslator.closeBlock(language);
    	
    	stringsToWrite.put(linkClass,strings);
    	System.out.println("CustomTypesGenerator: start to generate class named "+ path+className);
    }
    
    public static String getCurrentClassLink(int language){
    	String className = (Generator.isClient? "CustomTypes":"CustomTypes"+((Generator.encoding == Generator.STRING)? "4StringEncoding":""));
    	String currentClass = Generator.getPath(language,false)+className;
    	return currentClass;
    }
    
    public static String startToWriteEncodeDataMethod(int language, int encoding) {	
    	String encodeData = "";
    	if (encoding == Generator.STRING){
       		encodeData += "\t"+LanguageTranslator.addGlobalVar(language,true,false,Generator.isClient,true,"","String","encoding")+"\r\r"; 
       		//encodeData += "\t private static var encoding:String = \"\";\r\r";
       		
       		String[][] params = {{"Hashtable","h"}};
	       	if (Generator.isClient) encodeData += "\t"+LanguageTranslator.writeFunctionDeclaration(language,true,true,false,false,"String","encodeData",params,false,null)+"\r";
	       	else encodeData += "\t"+LanguageTranslator.writeFunctionDeclaration(language,true,false,false,false,"String","encodeData",params,false,null)+"\r";
	       	//encodeData += "\t public static function encodeData(h:Hashtable):String {\r";
        
	       	// Add a version number to be compliant with B4J binary encoding/decoding
       		String[] vars = {"\"\"+"+Generator.ENCODING_VERSION};
       		String value = LanguageTranslator.callMethod(language,false,"","addEncodingSpacer",vars);
       		encodeData += "\t\t"+LanguageTranslator.affectValueToVar(language,"encoding",value) +"\r";
       		//encodeData += "\t\t encoding = addEncodingSpacer(\""+ENCODING_VERSION+"\");\r"; 
       		
       	  	String[] vars1 = {"\"\"+"+LanguageTranslator.callMethod(language,true,"h","size",null)};
	       	String[] itemsToAdd1 = {LanguageTranslator.callMethod(language,false,"","addEncodingSpacer",vars1)};
	       	encodeData += "\t\t"+LanguageTranslator.add(language,true,"String","encoding",itemsToAdd1)+"\r";  
	       	//encodeData += "\t\t encoding += addEncodingSpacer(h.getSize());\r";
       	}
       	else if (encoding == Generator.BINARY){  		
       		// start to build root method of the Encoder : encodeData
       		String[][] params = {{"Hashtable","h"},{"DataOutputStream","dos"}};
       		String[] exceptions = {"Exception"};
       		if (Generator.isClient) encodeData += "\t"+LanguageTranslator.writeFunctionDeclaration(language,true,true,false,true,"","encodeData",params,true,exceptions)+"\r";
	       	else encodeData += "\t"+LanguageTranslator.writeFunctionDeclaration(language,true,false,false,true,"","encodeData",params,true,exceptions)+"\r";
	      
       		//encodeData += "\t public void encodeData(Hashtable h, DataOutputStream dos) throws Exception {\r";

            // Add a version number to be compliant with B4J binary encoding/decoding
            encodeData += "\t\t"+LanguageTranslator.writeBinaryType(language,"short",""+Generator.ENCODING_VERSION,"dos")+"\r";
       		//encodeData += "\t\t dos.writeShort("+ENCODING_VERSION+");\r";
            encodeData += "\t\t"+LanguageTranslator.writeBinaryType(
            							language,
            							"short",
            							LanguageTranslator.callMethod(language,true,"h","size",null),
            							"dos")+"\r";				
            //encodeData += "\t\t dos.writeShort(h.size());\r";
       	}
       		
     
       	encodeData += "\t\t"+LanguageTranslator.addLocalVar(
       						language,
            				true,
            				LanguageTranslator.callMethod(language,true,"h","size",null)
            				,"int"
            				,"nbSend"
            				)+"\r";
            //encodeData += "\t\t var nbSend = h.getSize();\r";
            
       	encodeData += "\t\t"+LanguageTranslator.forLoopDeclaration(language,"i","0","nbSend")+"\r";
       	//encodeData += "\t\t for(var i=0; i<nbSend; i++){\r";

       	if (Generator.objectsApproachMode){
	       	String[] vars2 = {"\"\"+i"};
	       	encodeData += "\t\t\t"+LanguageTranslator.addLocalVar(
	       					language,
	       					true,
            				LanguageTranslator.callMethod(FLASH,true,"h","get",vars2)
            				,"Object"
            				,"o"
            				)+"\r";
	       	//encodeData += "\t\t\t var o  = h.get(\"\"+i);\r";
	    }
        if (Generator.hashtablesApproachMode){
	       	String[] vars2 = {"\"\"+i"};
	       	encodeData += "\t\t\t"+LanguageTranslator.addLocalVar(
	       					language,
            				true,
            				LanguageTranslator.cast(language,"Hashtable",LanguageTranslator.callMethod(language,true,"h","get",vars2))
            				,"Hashtable"
            				,"hs"
            				)+"\r";
	       	//encodeData += "\t\t\t var hs:Hashtable  = h.get(\"\"+i);\r";
	    }
        
        if (Generator.objectsApproachMode && Generator.isCompactMode){
         	String init1 = LanguageTranslator.callMethod(language,true,LanguageTranslator.cast(language,Generator.destinationClassName,"o"),"getType",null);
        	encodeData += "\t\t\t"+LanguageTranslator.addLocalVar(language, true, init1,"byte","type")+"\r";
        }
        
        return encodeData;
    }
   
    
    public static void continueToWriteEncodeDataMethod(int language, int encoding, String typeNameFinal, String typeName, String typeClassName) {	
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
    	//String encodeData = strings[5];
    	                            
    	if (Generator.objectsApproachMode){
         	if (Generator.isCompactMode){        
             	String[] itemsToEquals = {"type", typeNameFinal};
             	strings[5] += "\t\t\t"+LanguageTranslator.ifDeclaration(language, LanguageTranslator.testEqualType(language,true,"int",itemsToEquals))+"\r";
                 //encodeData += "\t\t\t if ((("+destClassName+") o).getType() == type){\r";   
             }
             else{
            	 strings[5] += "\t\t\t"+LanguageTranslator.ifDeclaration(language,
	                						LanguageTranslator.testInstanceOfType(language,"o", typeClassName)
	                						)+"\r";
	                //encodeData += "\t\t\t if (o instanceof "+typeClassName+"){\r";
             }
         }
    	
         if (Generator.hashtablesApproachMode){
             String[] params3 = {"\"type\""};
             String[] items21 = {
             					LanguageTranslator.callMethod(language,true,"hs","get",params3),
             					"\""+typeClassName+"\""
             					};
             String cond1 = LanguageTranslator.testEqualType(language, true, typeClassName,items21);
             strings[5] += "\t\t\t"+LanguageTranslator.ifDeclaration(language,cond1)+"\r";
             //encodeData += "\t\t\t if (hs.get(\"type\") === \""+typeClassName+"\"){\r";
         }
             	
         if (encoding == Generator.STRING){
	            String[] vars11 = {"\"\"+"+ typeNameFinal};//(isCompactMode? typeNameFinal:(""+indexType))};//"\"\"+"+LanguageTranslator.callMethod(language,true,"h","size",null)};
	     	    String[] itemsToAdd11 = {LanguageTranslator.callMethod(language,false,"","addEncodingSpacer",vars11)};
	     	    strings[5] += "\t\t\t\t"+LanguageTranslator.add(language,true,"String","encoding",itemsToAdd11)+"\r";
	            //encodeData += "\t\t\t\t encoding += addEncodingSpacer(\""+indexType+"\");\r"; 
         } else if (encoding == Generator.BINARY){
        	  	strings[5] += "\t\t\t\t"+LanguageTranslator.writeBinaryType(language,"byte",typeNameFinal,"dos")+"\r";
	            //encodeData += "\t\t\t\t dos.writeByte("+indexType+");\r";
         }
             	
  	    if (Generator.objectsApproachMode){
  	    	String type1 = typeClassName;
	        	if (Generator.isCompactMode) type1 = Generator.destinationClassName;
  	        
	        	if (encoding == Generator.STRING){
  	        	String[] params21 = {LanguageTranslator.cast(language,type1,"o")};
  	        	strings[5] += "\t\t\t\t"+LanguageTranslator.callMethod(language,false,"","encode"+typeName,params21)+LanguageTranslator.endOfLineMarker(language)+"\r";
  	        	//encodeData += "\t\t\t\t encode"+typeName+"("+typeClassName+"(o));\r";
  	     	}
  	        else if (encoding == Generator.BINARY){
  	        	String[] params21 = {"dos",LanguageTranslator.cast(language,""+type1,"o")};
  	        	strings[5] += "\t\t\t\t"+LanguageTranslator.callMethod(language,false,"","encode"+typeName,params21)+LanguageTranslator.endOfLineMarker(language)+"\r";
	     	      	//encodeData += "\t\t\t\t encode"+typeName+"(dos,("+typeClassName+") o);\r";
  	       	}
  	   	}
	    if (Generator.hashtablesApproachMode){
	           	if (encoding == Generator.STRING){
		         	String[] params21 = {"hs"};
		         	strings[5] += "\t\t\t\t"+LanguageTranslator.callMethod(language,false,"","encode"+typeName,params21)+LanguageTranslator.endOfLineMarker(language)+"\r";
		         	//encodeData += "\t\t\t\t encode"+typeName+"(hs);\r";
	          	}
	           	else if (encoding == Generator.BINARY){
	             	String[] params21 = {"dos","hs"};
	             	strings[5] += "\t\t\t\t"+LanguageTranslator.callMethod(language,false,"","encode"+typeName,params21)+LanguageTranslator.endOfLineMarker(language)+"\r";
		           	//encodeData += "\t\t\t\t encodeData += "\t\t\t\t encode"+typeName+"(dos, hs);\r";
	       		}
	        }
	                
	    strings[5] += "\t\t\t"+LanguageTranslator.closeBlock(language)+"\r";
	    //encodeData += "\t\t\t }\r";

        //strings[5] = encodeData;
        //stringsToWrite.put(currentCustomTypes,strings);
    }
    
    
    public static String startToWriteDecodeDataMethod(int language, int encoding) {	
    	String decodeData = "";
    	String init = "";
       	String[] vars3 = {"encodedString"};
       	if (encoding == Generator.STRING){
	       	// start to build root method of the Decoder : decodeData
	       	decodeData += "\t"+LanguageTranslator.addGlobalVar(language,
	            			false,false,Generator.isClient,true,"0","int","indexTreatment")+"\r";
	       	//decodeData += "\t private static var indexTreatment = 0;\r\r";
	            
	       	String[][] params2 = {{"String","encodedString"}};
	        if (Generator.isClient)decodeData += "\t"+LanguageTranslator.writeFunctionDeclaration(language,
	            						true,true,false,false,"Hashtable","decodeData",params2,false,null)+"\r";  
	        else decodeData += "\t"+LanguageTranslator.writeFunctionDeclaration(language,
           						true,false,false,false,"Hashtable","decodeData",params2,false,null)+"\r"; 
	       	//decodeData += "\t public static function decodeData(encodedString:String):Hashtable {\r";
	            
	       	// read the version number to be compliant with B4J binary encoding/decoding
	       	decodeData += "\t\t"+LanguageTranslator.affectValueToVar(language,"indexTreatment","0")+"\r";
	       	//decodeData += "\t\t indexTreatment = 0;\r";
	       	
	       	decodeData += "\t\t"+LanguageTranslator.callMethod(language,false,"","getNextParam",vars3)+LanguageTranslator.endOfLineMarker(FLASH)+"\r";  
	       	//decodeData += "\t\t var encoding_version = getNextParam(encodedString); // ENCODING_VERSION\r";  
	        decodeData += "\t\t"+LanguageTranslator.addLocalVar(language,true,"","Hashtable","h")+"\r";
	    	//decodeData += "\t\t var h:Hashtable = new Hashtable();\r";
       	
	    	init = LanguageTranslator.callMethod(language,false,"","getNextParam",vars3);
	       	decodeData += "\t\t"+LanguageTranslator.addLocalVar(language,
	            				true,
	            				LanguageTranslator.parseStringType(language,"short",init)
	            				,"short"
	            				,"nb")+"\r"; 
	       	//decodeData += "\t\t var nb = parseInt(getNextParam(encodedString));\r";          
	      
       	} else if (encoding == Generator.BINARY){
       		//start to build root method of the Decoder : decodeData
	       	String[][] params2 = {{"DataInputStream","dis"}};
	       	String[] exceptions = {"Exception"};
	        if (Generator.isClient)decodeData += "\t"+LanguageTranslator.writeFunctionDeclaration(language,
	            						true,true,false,false,"Hashtable","decodeData",params2,true,exceptions)+"\r"; 
	        else decodeData += "\t"+LanguageTranslator.writeFunctionDeclaration(language,
           						true,false,false,false,"Hashtable","decodeData",params2,true,exceptions)+"\r";  
       		//decodeData += "\t public Hashtable decodeData(DataInputStream dis) throws Exception {\r";
            
            // read the version number to be compliant with B4J binary encoding/decoding
            decodeData += "\t\t"+LanguageTranslator.readBinaryType(language,"short","dis")+LanguageTranslator.endOfLineMarker(language)+"\r";
	        //decodeData += "\t\t dis.readShort(); // ENCODING_VERSION\r";  
            decodeData += "\t\t"+LanguageTranslator.addLocalVar(language,true,"","Hashtable","h")+"\r";
	    	//decodeData += "\t\t var h:Hashtable = new Hashtable();\r";
	       	decodeData += "\t\t"+LanguageTranslator.addLocalVar(language,
	            				true,
	            				LanguageTranslator.readBinaryType(language,"short","dis")
	            				,"short"
	            				,"nb")+"\r"; 
            //decodeData += "\t\t short nb = dis.readShort();\r";          
       	}
	   
       
        decodeData += "\t\t"+LanguageTranslator.addLocalVar(language,true,"nb","int","nbObj")+"\r";
       	//decodeData += "\t\t var nbObj = nb;\r";
       	String[] items1 = {"nb","0"};
       	decodeData += "\t\t"+LanguageTranslator.whileLoopDeclaration(language,
            					LanguageTranslator.testEqualType(language,false,"int",items1)
            					) +"\r";
       	//decodeData += "\t\t while(nb!=0){\r";
       	String[] items2 = {"nbObj","nb"};
       	decodeData += "\t\t\t"+LanguageTranslator.addLocalVar(language,
            				true,
            				"\"\"+"+LanguageTranslator.addValueToValue(language,false,items2),
            				"string",
            				"key")+"\r";
       	//decodeData += "\t\t\t var key:String = \"\"+(nbObj - nb);\r"; 
       	
       	if (encoding == Generator.STRING){
	       	decodeData += "\t\t\t"+LanguageTranslator.addLocalVar(language,
	            				true,
	            				LanguageTranslator.parseStringType(language,"byte",init)
	            				,"byte"
	            				,"type")+"\r"; 
	       	//decodeData += "\t\t\t var type = parseInt(getNextParam(encodedString));\r";  
       	}else if (encoding == Generator.BINARY){
       		decodeData += "\t\t\t"+LanguageTranslator.addLocalVar(language,
    				true,
    				LanguageTranslator.readBinaryType(language,"byte","dis")
    				,"byte"
    				,"type")+"\r";  
       		//decodeData += "\t\t\t byte type = dis.readByte();\r";
       	}
       	
       	return decodeData;
    }
    
    
    public static void continueToWriteDecodeDataMethod(int language, int encoding, int indexType, boolean isFirstIf, String typeNameFinal, String typeName, String typeClassName) {	
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
    	//String decodeData = strings[6];
    	
    	String[] itemsToEquals = {"type",(Generator.isCompactMode? typeNameFinal:(""+indexType))};
    	String cond2 = LanguageTranslator.testEqualType(language,true,"int",itemsToEquals);
        if (isFirstIf) {isFirstIf = false; strings[6] += "\t\t\t"+LanguageTranslator.ifDeclaration(language,cond2)+"\r";}
        else strings[6] += "\t\t\t"+ LanguageTranslator.elseIfDeclaration(language,cond2)+"\r";
        //if (isFirstIf) {isFirstIf = false; decodeData+="\t\t\t ";}
        //else decodeData+="\t\t\t else ";
        //decodeData += "if (type ===" +" "+indexType+"){\r";
               
        if (Generator.objectsApproachMode){	
        	String type1 = typeClassName;
	        	if (Generator.isCompactMode) type1 = Generator.destinationClassName;
	        	
        	if (encoding == Generator.STRING){
             	String[] vars31 = {"encodedString"};
             	strings[6] += "\t\t\t\t"+LanguageTranslator.addLocalVar(language,
	     	            				true,
	     	            				LanguageTranslator.callMethod(language,false,"","decode"+typeName,vars31)
	     	            				,""+type1
	     	            				,"customType")+"\r"; 
	         	//decodeData += "\t\t\t\t var customType:"+typeClassName+" = decode"+typeName+"(encodedString);\r";
	                
         	} else if (encoding == Generator.BINARY){
	          	String[] vars31 = {"dis"};
	          	strings[6] += "\t\t\t\t"+LanguageTranslator.addLocalVar(language,
 	            				true,
 	            				LanguageTranslator.callMethod(language,false,"","decode"+typeName,vars31)
 	            				,""+type1
 	            				,"customType")+"\r"; 
	                	//decodeData += "\t\t\t\t "+typeClassName+" customType = decode"+typeName+"(dis);\r";
	        }
        	
        	String[] params22 = {"key","customType"};
        	strings[6] += "\t\t\t\t"+LanguageTranslator.callMethod(language,true,"h","put",params22)+LanguageTranslator.endOfLineMarker(language)+"\r";
     	  	//decodeData += "\t\t\t\t h.put(key,customType);\r";
        }
    	if (Generator.hashtablesApproachMode){
              	if (encoding == Generator.STRING){
                	String[] vars31 = {"encodedString"};
                	strings[6] += "\t\t\t\t"+LanguageTranslator.addLocalVar(language,
	     	            				true,
	     	            				LanguageTranslator.callMethod(language,false,"","decode"+typeName,vars31)
	     	            				,"Hashtable"
	     	            				,"hs")+"\r"; 
	             	//decodeData += "\t\t\t\t var hs:Hashtable = decode"+typeName+"(encodedString);\r";  
               	} else if (encoding == Generator.BINARY){
                	String[] vars31 = {"dis"};
                	strings[6] += "\t\t\t\t"+LanguageTranslator.addLocalVar(language,
     	            				true,
     	            				LanguageTranslator.callMethod(language,false,"","decode"+typeName,vars31)
     	            				,"Hashtable"
     	            				,"hs")+"\r"; 
                	//decodeData += "\t\t\t\t Hashtable hs = decode"+typeName+"(dis);\r";
               	}
                	
             	String[] params23 = {"key","hs"};
             	strings[6] += "\t\t\t\t"+LanguageTranslator.callMethod(language,true,"h","put",params23)+LanguageTranslator.endOfLineMarker(language)+"\r";
             	//decodeData += "\t\t\t\t h.put(key,hs);\r";
    	}
                
    	strings[6] += "\t\t\t"+LanguageTranslator.closeBlock(language)+"\r";
    	//decodeData += "\t\t\t }\r";
        
    	//strings[6] = decodeData;
        //stringsToWrite.put(currentCustomTypes,strings);
    }
    
    
    /*****************
    appel: 	obj> unicastRecipientEncoding(o.recipient);
    		hasht> unicastRecipientEncoding(h.get("recipient"));
    */
    	
    /*****************
    appel: 	obj> short recipient = unicastRecipientDecoding(dis);
    			 o = new Type(recipient)
    		hasht> h.put("recipient",unicastRecipientDecoding(dis));
    */
    	
    /*****************
    appel: 	obj> multicastRecipientsEncoding(dos,o.recipients);
    		hasht> h.put("recipients",multicastRecipientsDecoding(dos));
    */
    	
    /*****************
     appel: obj> short[] recipients = multicastRecipientsDecoding(dis);
     			 Type o = new Type(recipients);
    		hasht> Hashtable h = new Hashtable();
    			   h.put("recipients",multicastRecipientsDecoding(dis));
     */
    
    public static void addRecipients4MulticastComModelEncoding(int language,int encoding){
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
    	if (encoding == Generator.BINARY){
	    	if (Generator.objectsApproachMode) strings[7] += "\t\t"+LanguageTranslator.callMethod(language,false,"","multicastRecipientsEncoding",new String[]{"dos",LanguageTranslator.callMethod(language,true,"customType","getRecipients",null)})+LanguageTranslator.endOfLineMarker(language)+"\r";
	    	if (Generator.hashtablesApproachMode) strings[7] += "\t\t"+LanguageTranslator.callMethod(language,false,"","multicastRecipientsEncoding",new String[]{"dos",LanguageTranslator.cast(language, "short[]",LanguageTranslator.callMethod(language,true,"h","get",new String[]{"\"recipients\""}))})+LanguageTranslator.endOfLineMarker(language)+"\r";
    	}
    	else if (encoding == Generator.STRING){
    		if (Generator.objectsApproachMode) strings[7] += "\t\t"+LanguageTranslator.callMethod(language,false,"","multicastRecipientsEncoding",new String[]{LanguageTranslator.callMethod(language,true,"customType","getRecipients",null)})+LanguageTranslator.endOfLineMarker(language)+"\r";
	    	if (Generator.hashtablesApproachMode) strings[7] += "\t\t"+LanguageTranslator.callMethod(language,false,"","multicastRecipientsEncoding",new String[]{  LanguageTranslator.cast(language, "short[]",LanguageTranslator.callMethod(language,true,"h","get",new String[]{"\"recipients\""}))})+LanguageTranslator.endOfLineMarker(language)+"\r";
    	}
    }
    
    
    public static void addRecipients4MulticastComModelDecoding(int language,int encoding){
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
    	if (encoding == Generator.BINARY){
	    	if (Generator.objectsApproachMode) strings[8] += "\t\t"+LanguageTranslator.addLocalVar(language,true,LanguageTranslator.callMethod(language,false,"","multicastRecipientsDecoding",new String[]{"dis"}),LanguageTranslator.arrayDeclaration(language,"short",1),"recipients")+LanguageTranslator.endOfLineMarker(language)+"\r";
	    	if (Generator.hashtablesApproachMode) strings[8] += "\t\t"+LanguageTranslator.callMethod(language,true,"h","put",new String[]{"\"recipients\"", LanguageTranslator.callMethod(language,false,"","multicastRecipientsDecoding",new String[]{"dis"})})+LanguageTranslator.endOfLineMarker(language)+"\r";
    	}
    	else if (encoding == Generator.STRING){
    		if (Generator.objectsApproachMode) strings[8] += "\t\t"+LanguageTranslator.addLocalVar(language,true,LanguageTranslator.callMethod(language,false,"","multicastRecipientsDecoding",new String[]{"encodedString"}),LanguageTranslator.arrayDeclaration(language,"short",1),"recipients")+LanguageTranslator.endOfLineMarker(language)+"\r";
	    	if (Generator.hashtablesApproachMode) strings[8] += "\t\t"+LanguageTranslator.callMethod(language,true,"h","put",new String[]{"\"recipients\"",LanguageTranslator.callMethod(language,false,"","multicastRecipientsDecoding",new String[]{"encodedString"})})+LanguageTranslator.endOfLineMarker(language)+"\r";
    	}
    	
    	strings[10] = writeMulticastCommunicationModelMethods(language,encoding);
    }
    
    
    
    public static void addRecipient4UnicastComModelEncoding(int language,int encoding){
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);

    	if (encoding == Generator.BINARY){
		    if (Generator.objectsApproachMode) strings[7] += "\t\t"+LanguageTranslator.callMethod(language,false,"","unicastRecipientEncoding",new String[]{"dos",LanguageTranslator.callMethod(language,true,"customType","getRecipient",null)})+LanguageTranslator.endOfLineMarker(language)+"\r";
		    if (Generator.hashtablesApproachMode) 
		    		strings[7] += "\t\t"+LanguageTranslator.callMethod(language,false,"","unicastRecipientEncoding",
		    				new String[]{"dos",LanguageTranslator.getObjectTypeValue(
		    							language,
		    							"short", LanguageTranslator.cast(language, "short", LanguageTranslator.callMethod(language,true,"h","get",new String[]{"\"recipient\""})))})+LanguageTranslator.endOfLineMarker(language)+"\r";
	    	}
    	else if (encoding == Generator.STRING){
    		if (Generator.objectsApproachMode) strings[7] += "\t\t"+LanguageTranslator.callMethod(language,false,"","unicastRecipientEncoding",new String[]{LanguageTranslator.callMethod(language,true,"customType","getRecipient",null)})+LanguageTranslator.endOfLineMarker(language)+"\r";
	    	if (Generator.hashtablesApproachMode) strings[7] += "\t\t"+LanguageTranslator.callMethod(language,false,"","unicastRecipientEncoding",new String[]{LanguageTranslator.cast(language, "short",LanguageTranslator.callMethod(language,true,"h","get",new String[]{"\"recipient\""}))})+LanguageTranslator.endOfLineMarker(language)+"\r";
    	}
    	
    	strings[11] = writeUnicastCommunicationModelMethods(language,encoding);
    }
    
    
    public static void addRecipient4UnicastComModelDecoding(int language,int encoding){
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
    	if (encoding == Generator.BINARY){
	    	if (Generator.objectsApproachMode) strings[8] += "\t\t"+LanguageTranslator.addLocalVar(language,true,LanguageTranslator.callMethod(language,false,"","unicastRecipientDecoding",new String[]{"dis"}),"short","recipient")+LanguageTranslator.endOfLineMarker(language)+"\r";
	    	if (Generator.hashtablesApproachMode) strings[8] += "\t\t"+LanguageTranslator.callMethod(language,true,"h","put",new String[]{"\"recipient\"",LanguageTranslator.callMethod(language,false,"","unicastRecipientDecoding",new String[]{"dis"})})+LanguageTranslator.endOfLineMarker(language)+"\r";
    	}
    	else if (encoding == Generator.STRING){
    		if (Generator.objectsApproachMode) strings[8] += "\t\t"+LanguageTranslator.addLocalVar(language,true,LanguageTranslator.callMethod(language,false,"","unicastRecipientDecoding",new String[]{"encodedString"}),"short","recipient")+LanguageTranslator.endOfLineMarker(language)+"\r";
	    	if (Generator.hashtablesApproachMode) strings[8] += "\t\t"+LanguageTranslator.callMethod(language,true,"h","put",new String[]{"\"recipient\"",LanguageTranslator.callMethod(language,false,"","unicastRecipientDecoding",new String[]{"encodedString"})})+LanguageTranslator.endOfLineMarker(language)+"\r";
    	}
    }
   
    
    public static void startToWriteEncodingMethods(int language, int encoding, String typeClassName, String typeName, String destinationClassName,int currentTypeCommunicationModel ) {	
    	System.out.println("CustomTypesGenerator: add method encode"+typeName);
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
    	//String encodeMethods = strings[7];
    	
    	if (Generator.objectsApproachMode){ 
    		String type1 = typeClassName;
     		if (Generator.isCompactMode) type1 = destinationClassName;
		        	
     		if (encoding == Generator.STRING){
            	String[][] attrs = {{type1,"customType"}};
            	strings[7]  += "\t"+LanguageTranslator.writeFunctionDeclaration(language,false,Generator.isClient,false,true,"","encode"+typeName, attrs, false, null)+"\r";
            	//encodeMethods+="\t public static function encode"+typeName+"(customType:"+typeClassName+"){ \r";
        	} else if (encoding == Generator.BINARY){
        		String[] exceptions = {"Exception"};
        		String[][] attrs = {{"DataOutputStream","dos"},{type1,"customType"}};
        		strings[7]  += "\t"+LanguageTranslator.writeFunctionDeclaration(language,false,Generator.isClient,false,true,"","encode"+typeName, attrs,true,exceptions)+"\r";
            	// encodeMethods+="\t public void encode"+typeName+"(DataOutputStream dos,"+typeClassName+" customType) throws Exception { \r";
        	}
     	}
    	if (Generator.hashtablesApproachMode){
        	if (encoding == Generator.STRING){
            	String[][] attrs = {{"Hashtable","h"}};
            	strings[7]  += "\t"+LanguageTranslator.writeFunctionDeclaration(language,false,Generator.isClient,false,true,"","encode"+typeName, attrs,false,null)+"\r";
            	//encodeMethods+="\t public static function encode"+typeName+"(h:Hashtable) { \r";
        	}else if (encoding == Generator.BINARY){
        		String[] exceptions = {"Exception"};
        		String[][] attrs = {{"DataOutputStream","dos"},{"Hashtable","h"}};
        		strings[7]  += "\t"+LanguageTranslator.writeFunctionDeclaration(language,false,Generator.isClient,false,true,"","encode"+typeName, attrs,true,exceptions)+"\r";
            	//encodeMethods+="\t public void encode"+typeName+"(DataOutputStream dos, Hashtable h) throws Exception { \r";
        	}
    	}
        
    	switch (currentTypeCommunicationModel){
	        case Generator.MULTICAST: 
	        	CustomTypesGenerator.addRecipients4MulticastComModelEncoding(language,encoding);
	        	break;
	        case Generator.UNICAST: 
	        	CustomTypesGenerator.addRecipient4UnicastComModelEncoding(language,encoding);
	        	break;

	        default: break;
    	}
        //strings[7] = encodeMethods;
        //stringsToWrite.put(currentCustomTypes,strings);
    }
    
    
    public static void continueToWriteEncodingMethods(int language, int encoding,  String elementType, String elementName) {	
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
    	//String encodeMethods = strings[7];
    	
        if (Generator.objectsApproachMode){
            String[] params25 = {LanguageTranslator.callMethod(language,true,"customType","get"+elementName,null)};
            strings[7] += "\t\t"+LanguageTranslator.callMethod(language,false,"","encode"+elementType,params25)+"\r";
            //encodeMethods+="\t\t encode"+elementType+"(customType.get"+elementName+"());\r";
        }
        if (Generator.hashtablesApproachMode){
            String[] params25 = {"dos","hs"};
            strings[7] += "\t\t"+LanguageTranslator.callMethod(language,false,"","encode"+elementType,params25)+"\r";
            //encodeMethods+="\t\t encode"+elementType+"(dos,hs);\r";
        }
        
        //strings[7] = encodeMethods;
        //stringsToWrite.put(currentCustomTypes,strings);
    }
    
    
    public static void generateEncodingType(int language, int encoding, String elementName, String type) {
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
    	//String encodeMethods = strings[7];
        
    	if (Generator.objectsApproachMode) {
        	if (encoding == Generator.STRING){
        		String toEncode = LanguageTranslator.callMethod(language,true,"customType","get"+elementName,null);
        		
        		//specific treatments to handle booleans on String
        		if (LanguageTranslator.getType(language,type).toLowerCase().equals("boolean")){
        			toEncode = "encodeBoolean("+toEncode+")";
        			Generator.isThereABooleanAttribute = true;
        		} else toEncode = "\"\"+"+toEncode;

	        	String[] params = {toEncode};
	        	String[] itemsToAdd = {LanguageTranslator.callMethod(language,false,"","addEncodingSpacer",params)};
	        	strings[7]  += "\t\t"+LanguageTranslator.add(language,true,"String","encoding",itemsToAdd)+"\r";  
	        	//toEncoder="\t\t encoding += addEncodingSpacer("+getFlashTypeEncoder(ClassGenerator.getFlashType(type),"customType.get"+elementName+"()")+");\r";
        	} else if (encoding == Generator.BINARY){
        		strings[7]  += "\t\t"+LanguageTranslator.writeBinaryType(
	        							language,
	        							type,
	        							LanguageTranslator.callMethod(language,true,"customType","get"+elementName,null),
	        							"dos")+"\r";
	        	// toEncoder="\t\t dos.write"+eType+"(customType.get"+elementName+"());\r";
        	}
        }
        if (Generator.hashtablesApproachMode) {
        	if (encoding == Generator.STRING){
        		String[] params2 = {"\""+elementName+"\""};
        		String toEncode = LanguageTranslator.callMethod(language,true,"h","get",params2);
        		
        		//specific treatments to handle booleans on String
        		if (LanguageTranslator.getType(language,type).toLowerCase().equals("boolean")){
        			toEncode = "encodeBoolean("+LanguageTranslator.getObjectTypeValue(
        										language, 
        										type,
        										LanguageTranslator.cast(language,type,toEncode))+")";
        			Generator.isThereABooleanAttribute = true;
        		} else toEncode = "\"\"+"+toEncode;

	        	String[] params = {toEncode};
	        	String[] itemsToAdd = {LanguageTranslator.callMethod(language,false,"","addEncodingSpacer",params)};
	        	strings[7]  += "\t\t"+LanguageTranslator.add(language,true,"String","encoding",itemsToAdd)+"\r";  
	        	//toEncoder="\t\t encoding += addEncodingSpacer("+getFlashTypeEncoder(ClassGenerator.getFlashType(type),"h.get(\""+elementName+"\")")+");\r";
        	} else if (encoding == Generator.BINARY){
	        	String[] params = {"\""+elementName+"\""};
        		String toWrite = LanguageTranslator.getObjectTypeValue(
							language,
							type,
							LanguageTranslator.cast(
								language, 
								type,
								LanguageTranslator.callMethod(language,true,"h","get",params))
							);
        		strings[7]  += "\t\t"+LanguageTranslator.writeBinaryType(
										language,
										type,
										toWrite,
										"dos")+"\r";
	        	//toEncoder="\t\t dos.write"+eType+"((("+getCastType(type)+") h.get(\""+elementName+"\"))."+type.toLowerCase()+"Value());\r";
        	}
        }
       
        //strings[7] = encodeMethods;
       // stringsToWrite.put(currentCustomTypes,strings);
	}
    
    
    public static void generateEncodingArrayType(int language, int encoding, boolean isACustomTypeArray, int dimension, int n, String type, boolean isLastElement) {
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
    	//String encodeMethods = strings[7];
        
    	String t = type.toLowerCase();
        String toEncoder="";
        String nTable= "table"+n;
        String nTableL= "table"+n+"_l";
        
	    if (Generator.objectsApproachMode){ 
	    	strings[7]  += "\t\t"+LanguageTranslator.addLocalVar(
	        							language,
	        							true,
	        							LanguageTranslator.callMethod(language,true,"customType","get"+upCaseFirstLetter(type),null),
	        							LanguageTranslator.arrayDeclaration(language,(Generator.isCompactMode? Generator.destinationClassName:type),dimension), 
	        							""+nTable
	        							)+"\r";
	    	//toEncoder+="\t\t var "+nTable+":Array = customType.get"+upCaseFirstLetter(type)+"();\r";
		}
	  	else if (Generator.hashtablesApproachMode){
	  		String[] params = {"\""+type+"\""};
	  		String arrayDeclaration = LanguageTranslator.arrayDeclaration(language,"Hashtable",dimension);    
	  		strings[7]  += "\t\t"+LanguageTranslator.addLocalVar(language,
						true,
						LanguageTranslator.cast(language,arrayDeclaration,LanguageTranslator.callMethod(language,true,"h","get",params)),
					    arrayDeclaration, 
						""+nTable
						)+"\r";
	  		//toEncoder+="\t\t var "+nTable+":Array = h.get(\""+type+"\");\r";
		}
        	
        int d = 0;
        
        while(d<dimension){
            String ind = indent(d);

            //String[] params = {"\""+type+"\""};
            strings[7]  += "\t\t"+ind+LanguageTranslator.addLocalVar(language,
						true,
						LanguageTranslator.arrayLength(language,nTable+tableLength(language,d)),
						"int", 
						nTableL+d
						)+"\r";
          	//toEncoder+="\t\t"+ind+" var "+nTableL+d+" = "+nTable+""+tableLength(d)+".length;\r";
	            
	     	if (encoding == Generator.STRING){
		     	String[] params1 = {"\"\"+"+nTableL+d};
		     	String[] itemsToAdd = {LanguageTranslator.callMethod(language,false,"","addEncodingSpacer", params1)};
		     	strings[7]  += "\t\t"+ind+LanguageTranslator.add(language,
		            					true,
		            					"String",
		            					"encoding",
		            					itemsToAdd)+"\r";  
		     	//toEncoder+="\t\t"+ind+" encoding += addEncodingSpacer("+nTableL+d+");\r";
	     	}
	      	else if (encoding == Generator.BINARY){
	      		strings[7] +="\t\t"+LanguageTranslator.writeBinaryType(language,"short",LanguageTranslator.castPrimitive(language,"short",""+nTableL+d),"dos")+"\r";
	        	//toEncoder+="\t\t"+ind+" dos.writeShort((short) "+nTableL+d+");\r";
	       	}
	        		            
	     	strings[7]  += "\t\t"+ind+LanguageTranslator.forLoopDeclaration(language,"i"+d,"0",""+nTableL+d)+"\r";
	       	//toEncoder+="\t\t"+ind+" for(var i"+d+"=0; i"+d+"<"+nTableL+d+"; i"+d+"++){\r";
            d++;
        }
        if (isACustomTypeArray){
        	if (encoding ==  Generator.STRING){
        		String[] params = {nTable+accessTable(language,dimension)};
        		strings[7]  += "\t\t"+indent(dimension)+LanguageTranslator.callMethod(language,false,"","encode"+upCaseFirstLetter(type),params)+LanguageTranslator.endOfLineMarker(language)+"\r";
            	//toEncoder+="\t\t"+indent(dimension)+" encode"+upCaseFirstLetter(type)+"("+nTable+accessTable(dimension)+");\r";
        	}
        	else if (encoding == Generator.BINARY){
        		String[] params = {"dos",nTable+accessTable(language,dimension)};
        		strings[7]  += "\t\t"+indent(dimension)+LanguageTranslator.callMethod(language,false,"","encode"+upCaseFirstLetter(type),params)+LanguageTranslator.endOfLineMarker(language)+"\r";
        		//toEncoder+="\t\t"+indent(dimension)+" encode"+upCaseFirstLetter(type)+"(dos,"+nTable+accessTable(dimension)+");\r";
        	}
        }
        else{
	        if (encoding == Generator.STRING){	
	        	strings[7]  +="\t\t"+indent(dimension)+" " +
		        			"encoding += addEncodingSpacer("+nTable+accessTable(language,dimension)+");\r"; 
	        } else if (encoding == Generator.BINARY){
	        	strings[7]  += "\t\t"+LanguageTranslator.writeBinaryType(language,type,""+nTable+accessTable(language,dimension),"dos")+"\r";
	        }
        }
        
        strings[7] += closeQuote(language, dimension);
        if (isLastElement) strings[7]  += "\r";

        //strings[7] = encodeMethods;
        //stringsToWrite.put(currentCustomTypes,strings);
	}
    
    
    public static void finishAnEncodingMethods(int language) {
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
		strings[7] += "\t"+LanguageTranslator.closeBlock(language)+"\r\r\r";
		//stringsToWrite.put(currentCustomTypes,strings);
	}
    
    
    public static void finishEncodeData(int language, int encoding) {
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
		strings[5] += "\t\t"+LanguageTranslator.closeBlock(language)+"\r";
        if (encoding == Generator.STRING) strings[5] += "\t\t"+LanguageTranslator.returnValue(language,"encoding")+"\r"; 
        strings[5] += "\t"+LanguageTranslator.closeBlock(language)+"\r";
    	//stringsToWrite.put(currentCustomTypes,strings);
	}
    
    
    public static void startToWriteDecodingMethods(int language, int encoding, String typeClassName, String typeNameFinal, String typeName, String destinationClassName, int currentTypeCommunicationModel) {	
    	System.out.println("CustomTypesGenerator: add method decode"+typeName);
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
    	//String decodeMethods = strings[8];
    	
    	 
        if (Generator.objectsApproachMode){
        	String type1 = typeClassName;
        	if (Generator.isCompactMode) type1 = destinationClassName;
		        	
        	if (encoding == Generator.STRING){
	           	String[][] attrs = {{"String","encodedString"}};
	           	strings[8] += "\t"+LanguageTranslator.writeFunctionDeclaration(language,true,Generator.isClient,false,false,type1,"decode"+typeName, attrs, false, null)+"\r";  
	          	//decodeMethods+="\t public static function decode"+typeName+"(encodedString:String):"+typeClassName+" {\r";
          	}else if (encoding == Generator.BINARY){
            	String[] exceptions = {"Exception"};
            	String[][] attrs = {{"DataInputStream","dis"}};
            	strings[8] += "\t"+LanguageTranslator.writeFunctionDeclaration(language,true,Generator.isClient,false,false,type1,"decode"+typeName, attrs, true, exceptions)+"\r";  
	        	// decodeMethods+="\t public "+typeClassName+" decode"+typeName+"(DataInputStream dis) throws Exception { \r";
          	}
        	
        	boolean isACommunicationModelSpecified = true;
        	String attribute = "";
        	switch (currentTypeCommunicationModel){
        		case Generator.MULTICAST: 
		        	CustomTypesGenerator.addRecipients4MulticastComModelDecoding(language,encoding);
		        	attribute = "recipients";
		        	break;
        		case Generator.UNICAST: 
		        	CustomTypesGenerator.addRecipient4UnicastComModelDecoding(language,encoding);
		        	attribute = "recipient";
		        	break;
		        	
        		default: 
        			isACommunicationModelSpecified = false; 
        			break;
        	}
        	
        	strings[8] += "\t\t"+LanguageTranslator.addLocalVar(
							language,
								true,
								isACommunicationModelSpecified? LanguageTranslator.newType(language,type1,attribute):""
								,type1
								,"customType")+"\r"+(Generator.isCompactMode? "":"\r");
          	//decodeMethods+="\t\t var customType:"+typeClassName+" = new "+typeClassName+"();\r";
        	
        	if (Generator.isCompactMode){
        		String[] items = {typeNameFinal};
        		strings[8] += "\t\t"+LanguageTranslator.callMethod(language,true,"customType","setType",items)+LanguageTranslator.endOfLineMarker(language)+"\r\r";   
        	}
        }
            	
      	if (Generator.hashtablesApproachMode){
         	if (encoding == Generator.STRING){
            	String[][] attrs = {{"String","encodedString"}};
            	strings[8] += "\t"+LanguageTranslator.writeFunctionDeclaration(language,true,Generator.isClient,false,false,"Hashtable","decode"+typeName, attrs,false,null)+"\r";
            	//decodeMethods+="\t public static function decode"+typeName+"(encodedString:String):Hashtable {\r";
        	}else if (encoding == Generator.BINARY){
            	String[] exceptions = {"Exception"};	
            	String[][] attrs = {{"DataInputStream","dis"}};
            	strings[8] += "\t"+LanguageTranslator.writeFunctionDeclaration(language,true,Generator.isClient,false,false,"Hashtable","decode"+typeName, attrs,true,exceptions)+"\r";
            	//decodeMethods+="\t public Hashtable decode"+typeName+"(DataInputStream dis) throws Exception { \r";
        	}
         	
         	strings[8] += "\t\t"+LanguageTranslator.addLocalVar(
					language,
					true,
					"",
					"Hashtable",
					"h")+"\r";
         			//decodeMethods+="\t\t var h:Hashtable = new Hashtable();\r";
         	
         	boolean isACommunicationModelSpecified = true;
         	String attribute = "";
         	switch (currentTypeCommunicationModel){
	    		case Generator.MULTICAST: 
		        	CustomTypesGenerator.addRecipients4MulticastComModelDecoding(language,encoding);
		        	attribute = "recipients";
		        	break;
	    		case Generator.UNICAST: 
		        	CustomTypesGenerator.addRecipient4UnicastComModelDecoding(language,encoding);
		        	attribute = "recipient";
		        	break;
	
	    		default: 
	    			isACommunicationModelSpecified = false; 
	    			break;
         	}
         	
         	//isACommunicationModelSpecified? LanguageTranslator.newType(language,typeClassName,attribute):
	            	
        	String[] params24 = {"\"type\"","\""+typeClassName+"\""};
        	strings[8] += "\t\t"+LanguageTranslator.callMethod(language,true,"h","put",params24)+LanguageTranslator.endOfLineMarker(language)+"\r\r";
        	//decodeMethods+="\t\t h.put(\"type\",\""+typeClassName+"\");\r";
      	}   	
        
        //strings[8] = decodeMethods;
        //stringsToWrite.put(currentCustomTypes,strings);
    }
    
    
    public static void continueToWriteDecodingMethods(int language, int encoding,  String elementType, String elementName) {	
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
    	//String decodeMethods = strings[8];
    	
        if (Generator.objectsApproachMode){
            String[] params25 = {LanguageTranslator.callMethod(language,true,"customType","get"+elementName,null)};
            strings[8] += "\t\t"+LanguageTranslator.callMethod(language,false,"","encode"+elementType,params25)+"\r";
            //encodeMethods+="\t\t encode"+elementType+"(customType.get"+elementName+"());\r";
        }
        if (Generator.hashtablesApproachMode){
            String[] params25 = {"dos","hs"};
            strings[8] += "\t\t"+LanguageTranslator.callMethod(language,false,"","encode"+elementType,params25)+"\r";
            //encodeMethods+="\t\t encode"+elementType+"(dos,hs);\r";
        }
        
        //strings[8] = decodeMethods;
        //stringsToWrite.put(currentCustomTypes,strings);
    }
    
    
    public static void generateDecodingType(int language, int encoding, String elementName, String type) {
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
    	//String decodeMethods = strings[8];
    	
		if (Generator.objectsApproachMode){
			if (encoding == Generator.STRING){
				String[] params1 = {"encodedString"};
    			LanguageTranslator.callMethod(language,false,"","getNextParam",params1);
        		String toDecode = LanguageTranslator.callMethod(language,false,"","getNextParam",params1);
    			
        		//specific treatments to handle booleans on String
        		if (LanguageTranslator.getType(language,type).toLowerCase().equals("boolean")){
        			toDecode = "decodeBoolean("+toDecode+")";
        			Generator.isThereABooleanAttribute = true;
        		}
    			
        		String[] params = {LanguageTranslator.parseStringType(language,type, toDecode)};
        		strings[8] += "\t\t"+LanguageTranslator.callMethod(language,true,"customType","set"+elementName, params)+LanguageTranslator.endOfLineMarker(language)+"\r"; 
    			//toDecoder = "\t\t customType.set"+elementName+"("+getFlashTypeDecoder(ClassGenerator.getFlashType(type),"getNextParam(encodedString)")+");\r";
			} else if (encoding == Generator.BINARY){
				String[] params = {LanguageTranslator.readBinaryType(language,type,"dis")};
				strings[8] += "\t\t"+LanguageTranslator.callMethod(language,true,"customType","set"+elementName, params)+LanguageTranslator.endOfLineMarker(language)+"\r"; 
    			//toDecoder = "\t\t customType.set"+elementName+"(dis.read";
			}
		}
    	if (Generator.hashtablesApproachMode){
    		if (encoding == Generator.STRING){
    			String[] params1 = {"encodedString"};
    			LanguageTranslator.callMethod(language,false,"","getNextParam",params1);
        		String toDecode = LanguageTranslator.callMethod(language,false,"","getNextParam",params1);
    			
        		//specific treatments to handle booleans on String
        		if (LanguageTranslator.getType(language,type).toLowerCase().equals("boolean")){
        			toDecode = "decodeBoolean("+toDecode+")";
        			Generator.isThereABooleanAttribute = true;
        		}

    			String[] params = {"\""+elementName+"\"",LanguageTranslator.newType(language, type,LanguageTranslator.parseStringType(language,type,toDecode))};
    			strings[8] += "\t\t"+LanguageTranslator.callMethod(language,true,"h","put", params)+LanguageTranslator.endOfLineMarker(language)+"\r"; 
        		//toDecoder = "\t\t h.put(\""+elementName+"\","+getFlashTypeDecoder(ClassGenerator.getFlashType(type),"getNextParam(encodedString)")+");\r";
    		} else if (encoding == Generator.BINARY){
    			String[] params = {"\""+elementName+"\"",LanguageTranslator.newType(language, type, LanguageTranslator.readBinaryType(language,type,"dis"))};
    			strings[8] += "\t\t"+LanguageTranslator.callMethod(language,true,"h","put", params)+LanguageTranslator.endOfLineMarker(language)+"\r"; 
    			//toDecoder = "\t\t h.put(\""+elementName+"\", new "+getCastType(type)+"(dis.read";
    		}
    	}
    	
        //strings[8] = decodeMethods;
        //stringsToWrite.put(currentCustomTypes,strings);
	}
    
    
    public static void generateDecodingArrayType(int language, int encoding, boolean isACustomTypeArray, int dimension, int n, String type, boolean isLastElement) {
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
    	//String decodeMethods = strings[8];
    	
    	String t = type.toLowerCase();
        String toDecoder="";
        String nTable= "table"+n;
        String nTableL= "table"+n+"_l";

        boolean noLowerCase = false;
        if (type.equals("String")||isACustomTypeArray) noLowerCase = true;

        int d = 0;
        
        while(d<dimension){
        	String ind = indent(d);
        	if (encoding == Generator.STRING){
        		String[] params = {"encodedString"};
        		String parse = LanguageTranslator.callMethod(language,false,"","getNextParam",params);
        		strings[8]+= "\t\t"+ind+LanguageTranslator.addLocalVar(
	            							language,
	            							true,
	            							LanguageTranslator.parseStringType(language,"int",parse),
	            							"int",
	            							nTableL+d
	            			)+"\r";  
        		//toDecoder+="\t\t"+ind+" var "+nTableL+d+" = parseInt(getNextParam(encodedString));\r";
        	} else if (encoding == Generator.BINARY){
        		strings[8]+= "\t\t"+ind+LanguageTranslator.addLocalVar(
									language,
									true,
									LanguageTranslator.readBinaryType(language,"short","dis"),
									"short",
									nTableL+d
        		 				)+"\r";  
        	}
   
	        String init = "";
	        if (Generator.objectsApproachMode){ 
			    if (noLowerCase) init = LanguageTranslator.arrayInit(language,(Generator.isCompactMode? Generator.destinationClassName:type),dimension,nTableL+d);
			    else init = LanguageTranslator.arrayInit(language,t,dimension,nTableL+d);
			    String init1 = LanguageTranslator.getAttributesInitiator(language,init);
			           
			    strings[8]+= "\t\t"+ind+LanguageTranslator.addLocalVar(
			            		language,
								true,
								init1,
								LanguageTranslator.arrayDeclaration(language,(Generator.isCompactMode? Generator.destinationClassName:type),dimension)
								,""+nTable)+"\r";
			    //toDecoder+="\t\t"+ind+" var "+nTable+":Array";
			    //if (noLowerCase)
			    //	toDecoder+=initTable(nTableL,dimension,d,type);
			    //else toDecoder+=initTable(nTableL,dimension,d,t);
			    //toDecoder+=";\r";
	 
			    strings[8]+= "\t\t"+ind+LanguageTranslator.forLoopDeclaration(language,"i"+d,"0",""+(nTableL+d))+"\r";
		        //toDecoder+="\t\t"+ind+" for(var i"+d+"=0; i"+d+"<"+nTableL+d+"; i"+d+"++){\r";
	        }
	        if (Generator.hashtablesApproachMode){
	        	init = LanguageTranslator.arrayInit(language,"Hashtable",dimension,nTableL+d);
			    String init1 = LanguageTranslator.getAttributesInitiator(language,init);
			           
			    strings[8]+= "\t\t"+ind+LanguageTranslator.addLocalVar(
			            		language,
								true,
								init1,
								LanguageTranslator.arrayDeclaration(language,"Hashtable",dimension)
								,""+nTable)+"\r";
			    //toDecoder+="\t\t"+ind+" var "+nTable+":Array";
			    //if (noLowerCase)
			    //	toDecoder+=initTable(nTableL,dimension,d,type);
			    //else toDecoder+=initTable(nTableL,dimension,d,t);
			    //toDecoder+=";\r";
	 
			    strings[8]+= "\t\t"+ind+LanguageTranslator.forLoopDeclaration(language,"i"+d,"0",""+(nTableL+d))+"\r";
		        //toDecoder+="\t\t"+ind+" for(var i"+d+"=0; i"+d+"<"+nTableL+d+"; i"+d+"++){\r";
	        }
	        d++;
        }
        
        if (isACustomTypeArray){
        	if (encoding == Generator.STRING){
	        	String[] params = {"encodedString"};
	        	strings[8] += "\t\t"+indent(dimension)+LanguageTranslator.affectValueToVar(
	                		 			language,
	                		 			nTable+accessTable(language,dimension),
	                		 			LanguageTranslator.callMethod(language,false,"","decode"+upCaseFirstLetter(type),params)
	                		 			)+"\r"; 
	        	//toDecoder+="\t\t"+indent(dimension)+" "+nTable+accessTable(FLASH,dimension)+"= decode"+upCaseFirstLetter(type)+"(encodedString);\r";
        	}
        	else if (encoding == Generator.BINARY){
        		String[] params = {"dis"};
        		strings[8] += "\t\t"+indent(dimension)+LanguageTranslator.affectValueToVar(
	                		 			language,
	                		 			nTable+accessTable(language,dimension),
	                		 			LanguageTranslator.callMethod(language,false,"","decode"+upCaseFirstLetter(type),params)
	                		 			)+"\r"; 
	        		 //toDecoder+="\t\t"+indent(dimension)+" "+nTable+accessTable(FLASH,dimension)+"= decode"+upCaseFirstLetter(type)+"(encodedString);\r";
        	}
         }
         else{
        	if (encoding == Generator.STRING){
	            String[] params = {"encodedString"};
	            String parse = LanguageTranslator.callMethod(language,false,"","getNextParam",params);
	            strings[8] += "\t\t"+indent(dimension)+LanguageTranslator.affectValueToVar(
	                					language,
	                		 			nTable+accessTable(language,dimension),
	                		 			LanguageTranslator.parseStringType(language,type,parse)
	                		 			)+"\r";  
	            //toDecoder+="\t\t"+indent(dimension)+" "+nTable+accessTable(FLASH,dimension)+"= ";
	            //toDecoder+= getFlashTypeDecoder(type,"getNextParam(encodedString)")+";\r";	
        	}
        	else if (encoding == Generator.BINARY) {
        		strings[8] += "\t\t"+indent(dimension)+LanguageTranslator.affectValueToVar(
										        					language,
										        		 			nTable+accessTable(language,dimension),
										        		 			LanguageTranslator.readBinaryType(language,type,"dis")
									        		 			)+"\r";
        	}
        }
             
        strings[8] += "\t"+closeQuote(language,dimension);
        
        if (Generator.objectsApproachMode){
        	String[] params = {nTable};
        	strings[8] += "\t\t"+LanguageTranslator.callMethod(language,true,"customType","set"+upCaseFirstLetter(type),params)+LanguageTranslator.endOfLineMarker(language)+"\r";
        	//toDecoder+="\t\t customType.set"+upCaseFirstLetter(type)+"("+nTable+");\r";
        }
        if (Generator.hashtablesApproachMode){
        	String[] params = {"\""+type+"\"",nTable};
        	strings[8] += "\t\t"+LanguageTranslator.callMethod(language,true,"h","put",params)+LanguageTranslator.endOfLineMarker(language)+"\r";
            //toDecoder+="\t\t h.put(\""+type+"\","+nTable+");\r";
        }
        
        if (isLastElement) strings[8] += "\r";
        
        //strings[8] = decodeMethods;
        //stringsToWrite.put(currentCustomTypes,strings);
	}

    
    public static void finishADecodingMethods(int language) {
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
    	if (Generator.objectsApproachMode) strings[8] += "\t\t"+LanguageTranslator.returnValue(language,"customType")+"\r";
        if (Generator.hashtablesApproachMode) strings[8] += "\t\t"+LanguageTranslator.returnValue(language,"h")+"\r";
        strings[8] += "\t"+LanguageTranslator.closeBlock(language)+"\r\r\r";
		//stringsToWrite.put(currentCustomTypes,strings);
	}

	
	public static void finishDecodeData(int language, int encoding) {
		String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
    	strings[6] += "\t\t\t"+LanguageTranslator.increment(language,false,"nb")+"\r";
    	strings[6] += "\t\t"+LanguageTranslator.closeBlock(language)+"\r";
    	strings[6] += "\t\t"+LanguageTranslator.returnValue(language,"h")+"\r";
    	strings[6] += "\t"+LanguageTranslator.closeBlock(language)+"\r";
    	//stringsToWrite.put(currentCustomTypes,strings);
	}
    
	
    public static void addImport(int language, String _import) {	
    	String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
    	strings[1] += LanguageTranslator.writeImport(language, _import)+"\r";
    	System.out.println("currentCustomTypes:"+currentCustomTypes);
		System.out.println("strings:"+strings);
    	stringsToWrite.put(currentCustomTypes,strings);
    }
    

    public static void finalize(int i) throws IOException{
    	String[] strings = (String[]) stringsToWrite.get(classNames.get(i));
    	int language = Integer.parseInt((String) classLanguages.get(i));
    	
    	if (!strings[0].equals("")) strings[0] += "\r";
    	if (!strings[1].equals("")) strings[1] += "\r";
    	if (!strings[2].equals("")) strings[2] += "\r";
    	if (!strings[3].equals("")) strings[3] = "\t// types\r"+strings[3]+"\r\r";
    	if (!strings[4].equals("")) strings[4] = "\t// constructor\r"+strings[4]+"\r\t"+LanguageTranslator.closeBlock(language)+"\r\r";
 
    	if (!strings[5].equals(""))
    		strings[5] = "\t// encodeData root method\r"+strings[5]+"\r\r";

    	if (!strings[6].equals(""))
    		strings[6] = "\t// decodeData rooy method\r"+strings[6]+"\r\r";

    	if (!strings[7].equals(""))
    		strings[7] = "\t// encoding method for each custom type\r"+strings[7]+"\r";
    	
    	if (!strings[8].equals(""))
    		strings[8] = "\t// decoding method for each custom type\r"+strings[8]+"\r";

    	if (!strings[9].equals(""))
    		strings[9] = "\t// string encoding required methods\r"+strings[9];

    	if (!strings[10].equals(""))
    		strings[10] = "\t// multicast communication methods\r"+strings[10];
    	
    	if (!strings[11].equals(""))
    		strings[11] = "\t// unicast communication methods\r"+strings[11];
    	
    	if (!strings[12].equals(""))
    		strings[12] += "\r";
    }
    
    public static void finalizeCustomTypesClasses() throws IOException{
	   	System.out.println("CustomTypesGenerator: finalize CustomTypes files...");
	   	for (int i=0; i<classNames.size(); i++) 
	   		finalize(i);
		System.out.println("CustomTypesGenerator: CustomTypes files finalized.");
   }

	public static void addGlobalVar(int language, String init, String type, String name) {
		String currentCustomTypes = getCurrentClassLink(language);
    	String[] strings = (String[]) stringsToWrite.get(currentCustomTypes);
    	strings[3] += "\t"+LanguageTranslator.addGlobalVar(language,true,true,Generator.isClient,true,init,type,name)+"\r";
    	stringsToWrite.put(currentCustomTypes,strings);
	}
	

    public static String writeUnicastCommunicationModelMethods(int language, int encoding){
    	String code = "";
	    String init = "";
	    String[][] params2;
	    
	    if (encoding == Generator.BINARY)
	    	params2 = new String[][]{{"DataInputStream","dis"}};

	    else /*(encoding == Generator.STRING)*/
	    	params2 = new String[][]{{"String","encodedString"}};


	    /************************************************************/
    	/** unicastRecipientEncoding(dos,recipient)                **/
    	/************************************************************/
	     
	    String[][] params3;
	    if (encoding == Generator.BINARY)  params3 = new String[][]{{"DataOutputStream","dos"},{"short","recipient"}};
	    else params3 = new String[][]{{"short","recipient"}};
	    
	    if (encoding == Generator.BINARY) {
	    	code +=  "\t"+LanguageTranslator.writeFunctionDeclaration(language,false,Generator.isClient,false,true,"","unicastRecipientEncoding",params3,true,new String[]{"Exception"})+"\r";
	    	code += "\t\t"+LanguageTranslator.writeBinaryType(language,"short","recipient","dos")+"\r";
	    }
	    else /*(encoding == Generator.STRING)*/{
	    	code +=  "\t"+LanguageTranslator.writeFunctionDeclaration(language,false,Generator.isClient,false,true,"","unicastRecipientEncoding",params3,false,null)+"\r";
	    	code += "\t\t"+LanguageTranslator.add(language,
					  								true,"String","encoding",
					  								new String[]{LanguageTranslator.callMethod(language,false,"","addEncodingSpacer",new String[]{LanguageTranslator.newType(language,"short","recipient")})}
	    										)+"\r";
	    }
	   
	    code +=  "\t"+LanguageTranslator.closeBlock(language)+"\r\r";
	    
	    /************************************************************/
    	/** unicastRecipientEncoding(dos,recipient)                **/
    	/************************************************************/

	    if (encoding == Generator.BINARY){ 
	    	code +=  "\t"+LanguageTranslator.writeFunctionDeclaration(language,false,Generator.isClient,false,false,"short","unicastRecipientDecoding",params2,true,new String[]{"Exception"})+"\r";
	    	code += "\t\t"+LanguageTranslator.returnValue(language,LanguageTranslator.readBinaryType(language,"short","dis"))+"\r";
	    }
	    else /*(encoding == Generator.STRING)*/{
	    	code +=  "\t"+LanguageTranslator.writeFunctionDeclaration(language,false,Generator.isClient,false,false,"short","unicastRecipientDecoding",params2,false,null)+"\r";
	    	code += "\t\t"+LanguageTranslator.returnValue(language,LanguageTranslator.parseStringType(language,"short",LanguageTranslator.callMethod(language,false,"","getNextParam", new String[]{"encodedString"})))+"\r";
		}
	   
	    code +=  "\t"+LanguageTranslator.closeBlock(language)+"\r\r";

    	return code;
    }
    
    
    public static String writeMulticastCommunicationModelMethods(int language, int encoding){
    	String code = "";
    	
    	/************************************************************/
    	/** multicastRecipientsDecoding(dos,short[])               **/
    	/************************************************************/
    	
    	if (encoding == Generator.BINARY){
	    	String[][] params = {{"DataOutputStream","dos"},{LanguageTranslator.arrayDeclaration(language,"short",1),"recipients"}};
		    code+= "\t"+LanguageTranslator.writeFunctionDeclaration(language,false,Generator.isClient,false,true,"","multicastRecipientsEncoding",params,true,new String[]{"Exception"})+"\r";
		    code+= "\t\t"+LanguageTranslator.writeBinaryType(language,"byte",LanguageTranslator.castPrimitive(language,"byte",LanguageTranslator.arrayLength(language,"recipients")),"dos")+"\r";

	    }
    	else if (encoding == Generator.STRING){
	    	String[][] params = {{LanguageTranslator.arrayDeclaration(language,"short",1),"recipients"}};
		    code+= "\t"+LanguageTranslator.writeFunctionDeclaration(language,false,Generator.isClient,false,true,"","multicastRecipientsEncoding",params,false,null)+"\r";
		    String newType =LanguageTranslator.newType(language,"int",LanguageTranslator.arrayLength(language,"recipients"));
		    code+= "\t\t"+LanguageTranslator.add(language, true, "String","encoding",new String[]{ LanguageTranslator.callMethod(language,false,"","addEncodingSpacer",new String[]{newType})})+"\r";
	    }
    	
	    code += "\t\t"+LanguageTranslator.forLoopDeclaration(language,"i","0",LanguageTranslator.arrayLength(language,"recipients"))+"\r";
	    
	    if (encoding == Generator.BINARY) code += "\t\t\t"+LanguageTranslator.writeBinaryType(language,"short",LanguageTranslator.arrayAccess(language,"recipients","i"),"dos")+"\r";
	    else if  (encoding == Generator.STRING){
	    	String[] params = {LanguageTranslator.newType(language,"short",LanguageTranslator.arrayAccess(language,"recipients","i"))};

	    	code += "\t\t\t"+LanguageTranslator.add(language,
	    			 							  true,"String","encoding",
	    			 							  new String[]{LanguageTranslator.callMethod(language,
	    			 									                                     false,"","addEncodingSpacer",
	    			 									                                     params)}
	    										 )+"\r";
	    }
	    
	    code += "\t\t"+LanguageTranslator.closeBlock(language)+"\r";
	    code += "\t"+LanguageTranslator.closeBlock(language)+"\r\r";
	    
	    /************************************************************/
    	/** short[] multicastRecipientsDecoding(dis)               **/
    	/************************************************************/
	    
	     String init = "";
	    String[][] params2;
	    String valueToInitArray = "";
	    if (encoding == Generator.BINARY){
	    	params2 = new String[][]{{"DataInputStream","dis"}};
	    	valueToInitArray = LanguageTranslator.readBinaryType(language,"byte","dis");
	    }
	    else /*(encoding == Generator.STRING)*/{
	    	params2 = new String[][]{{"String","encodedString"}};
	    	valueToInitArray = LanguageTranslator.parseStringType(language,"short",LanguageTranslator.callMethod(language,false,"","getNextParam",new String[]{"encodedString"}));	
	    }
	    
	    if (encoding == Generator.BINARY)
	    	code+= "\t"+LanguageTranslator.writeFunctionDeclaration(language,false,Generator.isClient,false,false,LanguageTranslator.arrayDeclaration(language,"short",1),
				"multicastRecipientsDecoding",params2,true,new String[]{"Exception"})+"\r";
	    else /*(encoding == Generator.STRING)*/
	    	code+= "\t"+LanguageTranslator.writeFunctionDeclaration(language,false,Generator.isClient,false,false,LanguageTranslator.arrayDeclaration(language,"short",1),
					"multicastRecipientsDecoding",params2,false,null)+"\r";
	    init = LanguageTranslator.getAttributesInitiator(
        									language,					
        									LanguageTranslator.arrayInit(
        											language,
        											"short",
        											1,
        											valueToInitArray
        									)
        						);
	    code += "\t\t"+LanguageTranslator.addLocalVar(language,true,init,LanguageTranslator.arrayDeclaration(language,"short",1),"recipients")+"\r";
	    code += "\t\t"+LanguageTranslator.forLoopDeclaration(language,"i","0",LanguageTranslator.arrayLength(language,"recipients"))+"\r";
	    
	    String valueToAffect = "";
	    if (encoding == Generator.BINARY) valueToAffect = LanguageTranslator.readBinaryType(language,"short","dis");
	    else /*(encoding == Generator.STRING)*/ valueToAffect = valueToInitArray;
	    
	    code += "\t\t\t"+LanguageTranslator.affectValueToVar(language, LanguageTranslator.arrayAccess(language,"recipients","i"),valueToAffect)+"\r";//(language,"short",LanguageTranslator.arrayAccess(language,"recipients","i"),"dos")+"\r";
	    code += "\t\t"+LanguageTranslator.closeBlock(language)+"\r";
	    code += "\t\t"+LanguageTranslator.returnValue(language,"recipients")+"\r";
	    code += "\t"+LanguageTranslator.closeBlock(language)+"\r\r";
	    
    	return code;
    }
    


	 /**
     * Generate string : ("}")*
     * 
     * @param dimension
     */
    private static String closeQuote(int language,int d) {
    	int i = 0;
        String ind = "";
        while (i<d){
            ind+=indent(d-i+1)+LanguageTranslator.closeBlock(language)+"\r";
            i++;
        }
        return ind;
    }


    /**
     * Generate string : ("[ii]")*
     * 
     * @param d dimension
     * @return
     */
    private static String accessTable(int language, int d) {
        int i = 0;
        String form = "";
        while (i<d){
            form+=LanguageTranslator.arrayAccess(language,"i"+i);
            i++;
        }
        return form;
    }


    /**
     * Generate string : "table"("[]")*
     * 
     * @param d dimension
     * @return
     */
    private static String tableLength(int language, int d) {
    	String form = "";
    	int i = 0;
        while (i<d){
        	form+=LanguageTranslator.arrayAccess(language,"i"+i);
	        i++;
	    }
        return form;
    }


    /**
     * Generate string : ("\t")*
     * 
     * @param d dimension
     * @return
     */
    private static String indent(int d) {
        int i = 0;
        String ind = "";
        while (i<d){
            ind+="\t";
            i++;
        }
        return ind;
    }


    /**
     * Generate table dimension code.
     * 
     * @param dimension
     * @return the string of the dimension code
     */
    public static String getDimension(int dimension) {
        String dim="";
        
        while(dimension>0){ 
            dim+="[]";
            dimension--;
        }
        return dim;
    }
    
    /**
     * @param s
     * @return
     */
    public static String upCaseFirstLetter(String s) {
        char[] cTab = s.toCharArray();
        cTab[0] = (cTab[0]+"").toUpperCase().charAt(0);
        return new String(cTab);
    }

	private static String addStringEncodingSpecificMethods(int language){  /*TODO: binaire*/
		String methods = "";
		
    	/****************************************/
    	/** Method addEncodingSpacer(Object o) **/
    	/****************************************/
    	
    	String[][] params = {{"Object","o"}};
    	methods+= "\t"+LanguageTranslator.writeFunctionDeclaration(language,false,Generator.isClient,false,false,"String","addEncodingSpacer",params,false,null)+"\r";
    	//methods+= "\t private static function addEncodingSpacer(o:Object):String {\r";

    	methods+= "\t\t"+LanguageTranslator.returnValue(language,"o+\""+Generator.FLASH_ENCODING_SPACER+"\"")+"\r";
    	//methods+= "\t\t return o+\""+ FLASH_ENCODING_SPACER +"\";\r";
    	methods+= "\t"+LanguageTranslator.closeBlock(language)+"\r\r";
    	//methods+= "\t }\r\r";
    	
    	/****************************************/
    	/** Method getNextParam(String code)   **/
    	/****************************************/
    	
    	String[][] params3 = {{"String","code"}};
    	methods+= "\t"+LanguageTranslator.writeFunctionDeclaration(language,false,Generator.isClient,false,false,"String","getNextParam",params3,false,null)+"\r";
    	//methods+= "\t private static function getNextParam(code:String):String{\r";
    		
    	methods+= "\t\t"+LanguageTranslator.addLocalVar(language,true,"","String","param")+"\r";
    	//methods+= "\t\t var param:String = \"\";\r";
    		
    	String[] params4 = {"indexTreatment"};
    	String init = LanguageTranslator.callMethod(language,true,"code","charAt",params4);
    	methods+= "\t\t"+LanguageTranslator.addLocalVar(language,true,init,"char","c")+"\r";
    	//methods+= "\t\t var c = code.charAt(indexTreatment);\r\r";
    		
    	String[] itemsToEquals2 = {"c","\'"+Generator.FLASH_ENCODING_SPACER+"\'"};
    	String[] itemsToTest2 = {"indexTreatment", LanguageTranslator.stringLength(language,"code")};
    	String[] itemsToTest = {LanguageTranslator.testSup(language,false,itemsToTest2),
    								LanguageTranslator.testEqualType(language,false,"char", itemsToEquals2)};
    	String cond1 = LanguageTranslator.testAnd(language,true,itemsToTest);
    	methods+= "\t\t"+LanguageTranslator.whileLoopDeclaration(language,cond1)+"\r";
    	//methods+= "\t\t while ((indexTreatment < length(code)) && c !== \""+ FLASH_ENCODING_SPACER +"\" ){\r";
    		
    	String[] paramsToAdd = {"c"};
    	methods+= "\t\t\t"+LanguageTranslator.add(language,true,"String","param", paramsToAdd)+"\r";
    	//methods+= "\t\t\t param += c;\r";
    		
    	methods+= "\t\t\t"+LanguageTranslator.increment(language,true,"indexTreatment")+"\r";
    	//methods+= "\t\t\t indexTreatment++;\r";
    		
    	methods+= "\t\t\t"+LanguageTranslator.ifDeclaration(language, LanguageTranslator.testSup(language,false,itemsToTest2))+"\r";
    	methods+= "\t\t\t\t"+LanguageTranslator.affectValueToVar(language,"c",LanguageTranslator.callMethod(language,true,"code","charAt",params4))+"\r";
    	methods+= "\t\t\t"+LanguageTranslator.closeBlock(language)+"\r";
    	//methods+= "\t\t\t if (indexTreatment < length(code)) c = code.charAt(indexTreatment);\r";
    		
    	methods+= "\t\t\t"+LanguageTranslator.elseDeclaration(language,"\t\t\t\t"+LanguageTranslator.loopInterrupt(language)) + "\r";
    	//methods+= "\t\t\t else break;\r";
    		
    	methods+= "\t\t\t"+LanguageTranslator.closeBlock(language)+"\r";
    	//methods+= "\t\t }\r";
    	
    	methods+= "\t\t"+LanguageTranslator.closeBlock(language)+"\r";
    	//methods+= "\t\t }\r";
    	
    	methods+= "\t\t"+LanguageTranslator.ifDeclaration(language, LanguageTranslator.testSup(language,false,itemsToTest2))+"\r";
    	methods+= "\t\t\t"+LanguageTranslator.increment(language,true,"indexTreatment")+LanguageTranslator.endOfLineMarker(language)+"\r";
    	//methods+= "\t\t if (indexTreatment < (length(code))) indexTreatment++; // \"pointeur\" apres |\r\r";
    		
    	methods+=  "\t\t"+LanguageTranslator.closeBlock(language)+"\r";
    	methods+=  "\t\t"+LanguageTranslator.returnValue(language,"param")+"\r";
    	//methods+= "\t\t return param;\r";

    	methods+=  "\t"+LanguageTranslator.closeBlock(language)+"\r";
    	//methods+= "\t }\r";
    	
    	if (Generator.isThereABooleanAttribute){
    		
	    	/****************************************/
	    	/** Method decodeBoolean(Object o)     **/
	    	/****************************************/
	    	
	    	String[][] params1 = {{"String","s"}};
	    	methods+= "\t"+LanguageTranslator.writeFunctionDeclaration(language,false,Generator.isClient,false,false,"boolean","decodeBoolean",params1,false,null)+"\r";
	    	//methods+= "\t private static function decodeBoolean(s:String):Boolean {\r";
	    		
	    	String[] itemsToEquals = {"s","\"0\""};
	    	String cond = LanguageTranslator.testEqualType(language,true,"String",itemsToEquals);
	    	methods+= "\t\t"+LanguageTranslator.ifDeclaration(language,cond)+"\r";
	    	methods+= "\t\t\t"+LanguageTranslator.returnValue(language,"false")+"\r";
	    	//methods+= "\t\t if (s === \"0\") return false;\r";
	    		
	    	methods+= "\t\t"+LanguageTranslator.closeBlock(language)+"\r";
	    	methods+= "\t\t"+LanguageTranslator.elseDeclaration(language,"\t\t\t"+LanguageTranslator.returnValue(language,"true"))+"\r";
	    	methods+= "\t\t"+LanguageTranslator.closeBlock(language)+"\r";
	    	//methods+= "\t\t else return true;\r";
	    		
	    	methods+= "\t"+LanguageTranslator.closeBlock(language)+"\r\r";
	    	//methods+= "\t }\r\r";
	    		
	    	
	    	/****************************************/
	    	/** Method encodeBoolean(Boolean b)    **/
	    	/****************************************/
	    	
	    	String[][] params2 = {{"boolean","b"}};
	    	methods+= "\t"+LanguageTranslator.writeFunctionDeclaration(language,false,Generator.isClient,false,false,"String","encodeBoolean",params2,false,null)+"\r";
	    	//methods+= "\t private static function encodeBoolean(b:Boolean):String {\r";
	    		
	    	methods+= "\t\t"+LanguageTranslator.ifDeclaration(language,"b")+"\r";
	    	methods+= "\t\t\t"+LanguageTranslator.returnValue(language,"\"1\"")+"\r";
	    	methods+= "\t\t"+LanguageTranslator.closeBlock(language)+"\r";
	    	//methods+= "\t\t if (b.valueOf()) return \"1\";\r";
	    		
	    	methods+= "\t\t"+LanguageTranslator.elseDeclaration(language, "\t\t\t"+LanguageTranslator.returnValue(language,"\"0\""))+"\r";
	    	methods+= "\t\t"+LanguageTranslator.closeBlock(language)+"\r";
	    	//methods+= "\t\t else return \"0\";\r";
	    	methods+= "\t"+LanguageTranslator.closeBlock(language)+"\r\r";
	    	//strings[8]+= "\t }\r\r"; 	
    	}
    	
    	return methods;
	}
}
