/**
 * @class HashTable
 * @version 0.1
 * @author <a href="http://www.lalex.com/blog/">LAlex</a>
 * @since 2004-10-8
 */

class Hashtable {
	
	private var _k:Array;
	private var _v:Array;
	
	public function Hashtable() {
		clear();
	}
	
	/**
	 * Find a value in an array and return it's position
	 * @param ar Array where to search
	 * @param v Value to search
	 * @return Found position. If not found, return -1
	 */
	private static function findValue(ar:Array, v):Number {
		var i:Number = ar.length;
		while(--i >= 0) {
			if (ar[i] == v) {
				break;
			}
		}
		return i;
	}
	
	/**
	 * Maps the specified key to the specified value in this hashtable
	 * @param k Hashtable key
	 * @param v Value
	 * @return Previous value of the specified key in this hashtable, or null if it did not have one
	 */
	public function put(k, v) {
		var i:Number = findValue(_k, k);
		if (i > -1) {
			var old = _v[i];
			_v[i] = v;
			return old;
		}
		_k.push(k);
		_v.push(v);
		return null;
	}
	
	/**
	 * Returns the value to which the specified key is mapped in this hashtable
	 * @param k A key in the hashtable
	 * @return The value to which the key is mapped in this hashtable, null if the key is not mapped to any value in this hashtable.
	 */
	public function get(k) {
		var i:Number = findValue(_k, k);
		if (i > -1) {
			return _v[i];
		}
		return null;
	}
	
	/**
	 * Removes the key (and its corresponding value) from this hashtable. This method does nothing if the key is not in the hashtable.
	 * @param k the key that needs to be removed
	 * @return the value to which the key had been mapped in this hashtable, or null if the key did not have a mapping.
	 */
	public function remove(k) {
		var i:Number = findValue(_k, k);
		if (i > -1) {
			var old = _v[i];
			_k.splice(i, 1);
			_v.splice(i, 1);
			return old;
		}
		return null;
	}
	
	/**
	 * Returns true if this Hashtable maps one or more keys to this value.
	 * @param v Value whose presence in this Hashtable is to be tested
	 * @return True if this map maps one or more keys to the specified value
	 */
	public function containsValue(v):Boolean {
		return findValue(_v, v) > -1;
	}

	/**
	 * Tests if the specified object is a key in this hashtable
	 * @param k Possible key
	 * @return True if the specified object is a key in this hashtable, false otherwise.
	 */
	public function containsKey(k):Boolean {
		return findValue(_k, k) > -1;
	}
	
	/**
	 * Returns the number of keys in this hashtable
	 * @return The number of keys in this hashtable
	 */
	public function size():Number {
		return _k.length;
	}
	
	/**
	 * Clears this hashtable so that it contains no keys.
	 */
	public function clear():Void {
		_k = new Array();
		_v = new Array();
	}
	
	/**
	 * Tests if this hashtable maps no keys to values.
	 * @return True if this hashtable maps no keys to values, false otherwise.
	 */
	public function isEmpty():Boolean {
		return _k.length > 0;
	}
	
	/**
	 * Returns an array of the keys in this hashtable.
	 * @return An array of the keys in this hashtable.
	 */
	public function getKeys():Array {
		var a:Array = new Array();
		var l:Number = _k.length;
		for (var i:Number = 0 ; i<l ; i++) {
			a[i] = _k[i];
		}
		return a;
	}
	
	/**
	 * Returns an array of the values in this hashtable.
	 * @return An array of the values in this hashtable.
	 */
	public function getValues():Array {
		var a:Array = new Array();
		var l:Number = _v.length;
		for (var i:Number = 0 ; i<l ; i++) {
			a[i] = _v[i];
		}
		return a;
	}
	
	/**
	 * Returns a string representation of this Hashtable object
	 * @return A string representation of this hashtable.
	 */
	public function toString():String {
		var st:String = "";
		var i:Number = _k.length;
		while (--i >= 0) {
			st = _k[i].toString() + " => " + _v[i].toString() +  (i < _k.length-1 ? "\n  " : " ") + st;
		}
		return "[ " + st + "]";
	}
	
}